import os
import sys
from pathlib import Path

import click
from bitmath import Byte, TiB
from boto3 import set_stream_logger
from click.core import Context

from NGPIris import HCPHandler


def add_trailing_slash(path: str) -> str:
    """
    Add a trailing slash ("/") to `path`.

    :param path: Arbitrary string
    :type path: str

    :return: Arbitrary string with `"/"` at the end
    :rtype: str
    """
    if not path.endswith("/"):
        path += "/"
    return path


def create_HCPHandler(context: Context) -> HCPHandler:
    """
    Returns a `HCPHandler` based on the given command `context`.

    :param context: The `click` context from the entered command
    :type context: Context

    :return: An `HCPHandler` instance based on the given command `context`
    :rtype: HCPHandler
    """
    if context.parent:
        parent_context = context.parent
    else:
        # Should never happen
        click.echo(
            (
                "Something went wrong with the subcommand and parent"
                "command relation"
            ),
            err=True,
        )
        sys.exit(1)

    credentials: str | None = parent_context.params.get("credentials")

    if credentials:
        hcp_credentials = credentials
    elif os.environ.get("NGPIRIS_CREDENTIALS_PATH", None):
        hcp_credentials = os.environ["NGPIRIS_CREDENTIALS_PATH"]
    else:
        endpoint: str = click.prompt(
            "Please enter your tenant endpoint",
        )

        aws_access_key_id: str = click.prompt(
            "Please enter your base64 hashed aws_access_key_id",
        )

        aws_secret_access_key: str = click.prompt(
            "Please enter your md5 hashed aws_secret_access_key",
            hide_input=True,
            confirmation_prompt=True,
        )

        hcp_credentials = {
            "endpoint": endpoint,
            "aws_access_key_id": aws_access_key_id,
            "aws_secret_access_key": aws_secret_access_key,
        }

    debug: bool | None = parent_context.params.get("debug")
    transfer_config: str | None = parent_context.params.get("transfer_config")
    if transfer_config:
        hcp_h = HCPHandler(hcp_credentials, custom_config_path=transfer_config)
    else:
        hcp_h = HCPHandler(hcp_credentials)

    if debug:
        set_stream_logger(name="")
        click.echo(hcp_h.transfer_config.__dict__)

    return hcp_h


def object_is_folder(object_path: str, hcp_h: HCPHandler) -> bool:
    """
    Predicate for checking if an HCP object is a folder or not.
    """
    return (object_path.endswith("/")) and (
        hcp_h.get_object(object_path)["ContentLength"] == 0
    )


def ensure_destination_dir(destination: str) -> Path:
    """
    Ensure that `destination` exists and return it as a `Path`.
    """
    dest_path = Path(destination)
    dest_path.mkdir(parents=True, exist_ok=True)
    return dest_path


def prompt_large_download() -> None:
    """
    Prompt for large downloads.
    """
    if not click.confirm(
        "WARNING: You are about to download more than 1 TB of data. "
        "Is this your intention?"
    ):
        sys.exit("\nAborting download")


def download_folder(
    source: str, destination_path: Path, ignore_warning: bool, hcp_h: HCPHandler
) -> None:
    """
    Helper function to `download` for downloading a folder.
    """
    prefix = "" if source == "/" else source

    cumulative_download_size = Byte(0)
    if not ignore_warning:
        click.echo("Computing download size...")
        for hcp_object in hcp_h.list_objects(prefix):
            hcp_object: dict
            cumulative_download_size += Byte(hcp_object["Size"])
            if cumulative_download_size >= TiB(1):
                prompt_large_download()

    hcp_h.download_folder(prefix, destination_path.as_posix())


def download_file(
    source: str,
    destination_path: Path,
    ignore_warning: bool,
    force: bool,
    hcp_h: HCPHandler,
) -> None:
    """
    Helper function to `download` for downloading a file.
    """
    check_size_and_ignore_warning_flag = (
        Byte(hcp_h.get_object(source)["ContentLength"]) >= TiB(1)
    ) and (not ignore_warning)
    if check_size_and_ignore_warning_flag:
        prompt_large_download()

    downloaded_source = Path(destination_path) / Path(source).name
    if downloaded_source.exists() and not force:
        sys.exit(
            "Object already exists. If you wish to overwrite the existing file,"
            " use the -f / --force option"
        )
    hcp_h.download_file(source, downloaded_source.as_posix())



from json import load
from pathlib import Path

from requests import Response, post
from urllib3 import disable_warnings

from NGPIris.hci.helpers import get_index_response, get_query_response
from NGPIris.parse_credentials import CredentialsHandler


class HCIHandler:
    """
    Class for handling HCI requests.
    """

    def __init__(
        self, credentials: str | dict[str, str], use_ssl: bool = False
    ) -> None:
        """
        Class for handling HCI requests.

        :param credentials:
            If `credentials` is a `str`, then it will be interpreted as a path
            to the JSON credentials file. If `credentials` is a `dict`, then a
            dictionary with the appropriate HCI credentials is expected:
            ```
            {
                "username" : "",
                "password" : "",
                "address" : "",
                "auth_port" : "",
                "api_port" : ""
            }
            ```
        :type credentials: str | dict[str, str]

        :param use_ssl: Boolean choice between using SSL, defaults to False
        :type use_ssl: bool, optional
        """
        if type(credentials) is str:
            credentials_handler = CredentialsHandler(credentials)
            self.hci = credentials_handler.hci

            self.username = self.hci["username"]
            self.password = self.hci["password"]
            self.address = self.hci["address"]
            self.auth_port = self.hci["auth_port"]
            self.api_port = self.hci["api_port"]
        elif type(credentials) is dict:
            self.username = credentials["username"]
            self.password = credentials["password"]
            self.address = credentials["address"]
            self.auth_port = credentials["auth_port"]
            self.api_port = credentials["api_port"]

        self.token = ""

        self.use_ssl = use_ssl

        if not self.use_ssl:
            disable_warnings()

    def request_token(self) -> None:
        """
        Request a token from the HCI, which is stored in the HCIHandler object.
        The token is used for every operation that needs to send a request to
        HCI.

        :raises ConnectionError:
            If there was a problem when requesting a token
        """
        url = "https://" + self.address + ":" + self.auth_port + "/auth/oauth/"
        data = {
            "grant_type": "password",
            "username": self.username,
            "password": self.password,
            "scope": "*",
            "client_secret": "hci-client",
            "client_id": "hci-client",
            "realm": "LOCAL",
        }
        try:
            response: Response = post(
                url, data=data, verify=self.use_ssl, timeout=15
            )
        except:  # noqa: E722  # pragma: no cover
            error_msg: str = (
                "The token request made at "
                + url
                + " failed. Please check your connection."
            )
            raise ConnectionError(error_msg) from None

        token: str = response.json()["access_token"]
        self.token = token

    def list_index_names(self) -> list[str]:
        """
        Retrieve a list of all index names.

        :return: A list of index names
        :rtype: list[str]
        """
        response: Response = get_index_response(
            self.address,
            self.api_port,
            self.token,
            self.use_ssl,
        )
        return [entry["name"] for entry in response.json()]

    def look_up_index(self, index_name: str) -> dict:
        """
        Look up index information in the form of a dictionary by submitting
        the index name. Will return an empty dictionary if no index was found.

        :param index_name: The index name
        :type index_name: str

        :return: A dictionary containing information about an index
        :rtype: dict
        """
        response: Response = get_index_response(
            self.address,
            self.api_port,
            self.token,
            self.use_ssl,
        )

        for entry in response.json():
            if entry["name"] == index_name:
                return dict(entry)

        return {}

    def raw_query(self, query_dict: dict[str, str | list | dict]) -> dict:
        """
        Make query to an HCI index, with a dictionary.

        :param query_dict: Dictionary consisting of the query
        :type query_dict: dict[str, str | list | dict]

        :return: Dictionary containing the raw query
        :rtype: dict
        """
        return dict(
            get_query_response(
                query_dict,
                self.address,
                self.api_port,
                self.token,
                self.use_ssl,
            ).json(),
        )

    def raw_query_from_JSON(self, query_path: str) -> dict:
        """
        Make query to an HCI index, with prewritten query in a JSON file.

        :param query_path: Path to the JSON file
        :type query_path: str

        :return: Dictionary containing the raw query
        :rtype: dict
        """
        with Path(query_path).open() as inp:
            return dict(
                get_query_response(
                    dict(load(inp)),
                    self.address,
                    self.api_port,
                    self.token,
                    self.use_ssl,
                ).json(),
            )

    def query(
        self,
        index_name: str,
        query_string: str = "",
        facets: list[str] = [],  # noqa: B006
    ) -> dict:
        """
        Make a query to the HCI based on the parameters of this method.

        :param index_name: Name of the index
        :type index_name: str

        :param query_string: The Solr query string. Defaults to the empty string
        :type query_string: str, optional

        :param facets:
            List of facets that should be returned included in the response.
            Defaults to []
        :type facets: list[str], optional

        :return: The response in the form of a dictionary
        :rtype: dict
        """
        facetRequests = [{"fieldName": facet} for facet in facets]
        return self.raw_query(
            {
                "indexName": index_name,
                "queryString": query_string,
                "facetRequests": facetRequests,
            },
        )



from NGPIris.parse_credentials import CredentialsHandler


def incomplete_creds(path: str) -> None:
    try:
        CredentialsHandler(path)
    except:
        assert True
    else:  # pragma: no cover
        assert False


def test_incomplete_creds_0() -> None:
    incomplete_creds("tests/data/incomplete_creds_0.json")


def test_incomplete_creds_1() -> None:
    incomplete_creds("tests/data/incomplete_creds_1.json")


def test_incomplete_creds_2() -> None:
    incomplete_creds("tests/data/incomplete_creds_2.json")


def test_incomplete_creds_3() -> None:
    incomplete_creds("tests/data/incomplete_creds_3.json")


def complete_creds(path: str) -> None:
    try:
        CredentialsHandler(path)
    except:  # pragma: no cover
        assert False
    else:
        assert True


def test_complete_creds_0() -> None:
    complete_creds("tests/data/complete_creds_0.json")


def test_complete_creds_1() -> None:
    complete_creds("tests/data/complete_creds_1.json")


def test_complete_creds_2() -> None:
    complete_creds("tests/data/complete_creds_2.json")



from json import load
from pathlib import Path
from typing import TypeAlias

Credentials: TypeAlias = dict[str, dict[str, str]]


class CredentialsHandler:
    """
    Class for handling credentials parsing.
    """

    def __init__(self, credentials_path: str) -> None:
        """
        Class for handling credentials to HCP and HCI.

        :param credentials_path: Path to the credentials JSON file
        :type credentials_path: str
        """
        self.hcp: dict[str, str] = {}
        self.hci: dict[str, str] = {}

        credentials: Credentials = parse_credentials(credentials_path)
        for key, value in credentials.items():
            setattr(self, key, value)


def all_fields_empty(key: str, credentials: Credentials) -> bool:
    """
    Predicate that checks if all fields in `credentials` with a given `key` are
    empty.

    :param key: Key for to be accessed in `credentials`
    :type key: str

    :param credentials: The credentials
    :type credentials: Credentials
    """
    return all(v == "" for v in credentials[key].values())


def check_empty_field(credentials: Credentials) -> None:
    """
    Makes the following checks:
     - All fields in the credentials file are empty
     - If fields under `hcp` and `hci` are only partially filled in
    If any of the above are true, then `RuntimeError` is raised

    :param credentials: Credentials to be checked
    :type credentials: Credentials

    :raise RuntimeError: If any check is true
    """  # noqa: D400, D415
    if all(all_fields_empty(k, credentials) for k in credentials):
        msg = (
            "Missing entries in all fields in the credentials file. "
            "Please enter your credentials in the credentials file"
        )
        raise RuntimeError(msg)
    empty_fields_per_entry: dict[str, list[str]] = {}
    for k1, d in credentials.items():
        # If all fields in *either* hci or hcp is empty then continue
        if all_fields_empty(k1, credentials):
            continue
        empty_fields: list[str] = []
        for k2, v in d.items():
            if v == "":
                empty_fields.append("\t- " + k2 + "\n")
        if empty_fields:
            empty_fields_per_entry[k1] = empty_fields

    all_empty_fields = []
    for entry, fields in empty_fields_per_entry.items():
        fields.insert(0, "- " + entry + ":\n")
        all_empty_fields.append("".join(fields))

    if all_empty_fields:
        msg = (
            "Missing fields for the following entries in the credentials"
            " file:\n" + "".join(all_empty_fields)
        )
        raise RuntimeError(msg)


def parse_credentials(credentials_path: str) -> Credentials:
    """
    Parse credentials at the given `credentials_path`.

    :param credentials_path: Path to credentials
    :type credentials_path: str

    :return: Parsed credentials
    :rtype: Credentials = dict[str, dict[str, str]]
    """
    with Path(credentials_path).open() as inp:
        credentials: Credentials = load(inp)
        check_empty_field(credentials)
        return credentials



import sys
from collections.abc import Callable
from pathlib import Path
from typing import ParamSpec, TypeVar

from NGPIris.hcp.exceptions import NoBucketMountedError


def create_access_control_policy(user_ID_permissions: dict[str, str]) -> dict:  # noqa: D103
    access_control_policy: dict[str, list] = {
        "Grants": [],
    }
    for user_ID, permission in user_ID_permissions.items():
        if permission not in [
            "FULL_CONTROL",
            "WRITE",
            "WRITE_ACP",
            "READ",
            "READ_ACP",
        ]:
            print("Invalid permission option:", permission)
            sys.exit()
        grantee = {
            "Grantee": {
                "ID": user_ID,
                "Type": "CanonicalUser",
            },
            "Permission": permission,
        }
        access_control_policy["Grants"].append(grantee)
    return access_control_policy


def raise_path_error(path: str) -> None:
    """
    Raise FileNotFoundError if the system path does not exist.

    :param path: Local system path
    :type path: str

    :raises FileNotFoundError: If `path` does not exist
    """
    if not Path(path).exists():
        raise FileNotFoundError('"' + path + '"' + " does not exist")


P = ParamSpec("P")
T = TypeVar("T")


def check_mounted(method: Callable[P, T]) -> Callable[P, T]:
    """
    Decorator for checking if a bucket is mounted. This is meant to be used by
    class methods, hence the possibly odd typing.

    :param method: An arbitrary class method of the `HCPHandler` class
    :type method: Callable[ParamSpec("P"), TypeVar("T")]

    :return: A decorated class method of the `HCPHandler` class
    :rtype: Callable[ParamSpec("P"), TypeVar("T")]
    """

    def check_if_mounted(*args: P.args, **kwargs: P.kwargs) -> T:
        self = args[0]
        if not self.bucket_name: # pyright: ignore[reportAttributeAccessIssue]
            msg = "No bucket is mounted"
            raise NoBucketMountedError(msg)
        return method(*args, **kwargs)

    return check_if_mounted



from json import dumps

from requests import Response, get, post


def get_index_response(
    address: str,
    api_port: str,
    token: str,
    use_ssl: bool,
) -> Response:
    """
    Retrieve the index response given the address, API port and token.

    :param address: The address where request is to be made
    :type address: str

    :param api_port: The API port at the given address
    :type api_port: str

    :param token: The HCI token
    :type token: str

    :param use_ssl: Boolean choice of using SSL
    :type use_ssl: bool

    :return: A response containing information about the index
    :rtype: requests.Response
    """
    url: str = "https://" + address + ":" + api_port + "/api/search/indexes/"
    headers: dict[str, str] = {
        "Accept": "application/json",
        "Authorization": "Bearer " + token,
    }

    response: Response = get(url, headers=headers, verify=use_ssl, timeout=15)

    response.raise_for_status()

    return response


def get_query_response(  # noqa: PLR0913
    query_dict: dict[str, str | list | dict],
    address: str,
    api_port: str,
    token: str,
    use_ssl: bool,
    path_extension: str = "",
) -> Response:
    """
    Retrieve the query response given the address, API port and token.

    :param query_dict: The query dictionary
    :type query_dict: dict[str, str | list | dict]

    :param address: The address where request is to be made
    :type address: str

    :param api_port: The API port at the given address
    :type api_port: str

    :param token: The HCI token
    :type token: str

    :param use_ssl: Boolean choice of using SSL
    :type use_ssl: bool

    :param path_extension:
        Possibly extend the request URL. Used for example when making SQL
        requests. Defaults to ""
    :type path_extension: str, optional

    :return: A response containing information about the query
    :rtype: requests.Response
    """
    if "indexName" not in query_dict:
        msg = "Field indexName is missing in the query dictionary"
        raise RuntimeError(msg)

    url: str = (
        "https://"
        + address
        + ":"
        + api_port
        + "/api/search/query/"
        + path_extension
    )
    query: dict[str, str | list | dict] = query_dict
    headers: dict[str, str] = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": "Bearer " + token,
    }
    response: Response = post(
        url, dumps(query), headers=headers, verify=use_ssl, timeout=15
    )

    response.raise_for_status()

    return response
